from abc import abstractmethod, ABC
from typing import Literal

from overrides import overrides, EnforceOverrides


class A(EnforceOverrides):
    def methoda(self, x=0):
        print(x)

    def methodb(self, x: int, /, y: str) -> str:
        return y * x


class Other:
    def foo(self):
        pass


def test_should_pass():
    class B(A):
        @overrides
        def methoda(self, y=1, **kwargs):
            print(y)
            super().methoda(**kwargs)


def test_should_also_pass():
    class B(A):
        @overrides
        def methoda(self, z=1, x=1, **kwargs):
            pass


class Abs(ABC):
    @abstractmethod
    def method(self, str: Literal["max", "min"]):
        pass


def test_literal_passes():
    class B(Abs):
        @overrides
        def method(self, str: Literal["max", "min"]):
            return

    class C(Abs):
        @overrides
        def method(self, str: Literal["max", "min", "half"]):
            return


def test_literal_failure():
    try:

        class D(Abs):
            @overrides
            def method(self, str: Literal["a", "b", "c"]):
                pass

        raise AssertionError("Should not go here")
    except TypeError:
        pass


def test_literal_failure_not_accepting_all():
    try:

        class D(Abs):
            @overrides
            def method(self, str: Literal["min"]):
                pass

        raise AssertionError("Should not go here")
    except TypeError:
        pass


def test_can_not_override_with_positional_only():
    try:

        class C(A):
            @overrides
            def methoda(self, x=0, /):
                pass

        raise AssertionError("Should not go here")
    except TypeError:
        pass


def test_can_override_positional_only():
    class PositionalOnly1(A):
        @overrides
        def methodb(self, x: int, /, y: str) -> str:
            return "OK"


def test_can_override_positional_only_with_new_name():
    class PositionalOnly2(A):
        @overrides
        def methodb(self, new_name_is_ok: int, /, y: str) -> str:
            return "OK2"


def test_can_not_override_positional_only_with_new_type():
    try:

        class PositionalOnly3(A):
            @overrides
            def methodb(self, x: str, /, y: str) -> str:
                return "NOPE"

        raise AssertionError("Should not go here")
    except TypeError:
        pass


def test_can_not_override_with_keyword_only():
    try:

        class C2(A):
            @overrides
            def methoda(self, *, x=0):
                pass

        raise AssertionError("Should not go here")
    except TypeError:
        pass


def test_multiple_inheritance():
    class Multi(A, Other):
        @overrides
        def methoda(self, y=2, **kwargs):
            pass

        @overrides
        def foo(self) -> int:
            pass



from typing import Tuple

from overrides import overrides


class Parent:
    def execute(self, x: int, y: str, z: bool, *args, **kwargs) -> Tuple[str, int]:
        return y, x


class Child(Parent):
    @overrides
    def execute(self, *args, **kwargs) -> Tuple[str, int]:
        return "moi", 1



import sys
from typing import Optional

import py


def pytest_ignore_collect(path: py.path.local, config: "Config") -> Optional[bool]:
    if sys.version_info[0] == 3 and sys.version_info[1] < 8:
        if str(path).endswith("__py38.py"):
            return True
    return None



from typing import Callable, Type
from overrides import overrides


class SuperClass:
    def method(self) -> int:
        return 2


def my_decorator(name: str) -> Callable:
    def func(cls: Type) -> Type:
        return cls

    return func


class MyClass:
    def __init__(self, name: str):
        self.my_name: str = name


def test_my_func() -> None:
    my_object = MyClass("Name accessed in decorator")

    @my_decorator(my_object.my_name)
    class SubClass(SuperClass):
        @overrides
        def method(self) -> int:
            return 1



from typing import Dict, Any, TypedDict

from overrides.typing_utils import issubtype


class MyTypedDict(TypedDict):
    foo: float
    bar: float


class Typed2(TypedDict):
    zoo: float


class Typed3(MyTypedDict, Typed2):
    pass


def test_typeddict_and_dict():
    assert issubtype(Typed3, Typed2)
    assert issubtype(Typed3, MyTypedDict)
    assert issubtype(MyTypedDict, Dict[str, Any])



from abc import ABCMeta


class EnforceOverridesMeta(ABCMeta):
    def __new__(mcls, name, bases, namespace, **kwargs):
        # Ignore any methods defined on the metaclass when enforcing overrides.
        for method in dir(mcls):
            if not method.startswith("__") and method != "mro":
                value = getattr(mcls, method)
                if not isinstance(value, (bool, str, int, float, tuple, list, dict)):
                    setattr(getattr(mcls, method), "__ignored__", True)

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)
        for name, value in namespace.items():
            # Actually checking the direct parent should be enough,
            # otherwise the error would have emerged during the parent class checking
            if name.startswith("__"):
                continue
            value = mcls.handle_special_value(value)
            is_override = getattr(value, "__override__", False)
            for base in bases:
                base_class_method = getattr(base, name, False)
                if (
                    not base_class_method
                    or not callable(base_class_method)
                    or getattr(base_class_method, "__ignored__", False)
                ):
                    continue
                assert (
                    is_override
                ), "Method %s overrides but does not have @overrides decorator" % (name)
                # `__finalized__` is added by `@final` decorator
                assert not getattr(base_class_method, "__finalized__", False), (
                    "Method %s is finalized in %s, it cannot be overridden"
                    % (base_class_method, base,)
                )
        return cls

    @staticmethod
    def handle_special_value(value):
        if isinstance(value, classmethod) or isinstance(value, staticmethod):
            value = value.__get__(None, dict)
        elif isinstance(value, property):
            value = value.fget
        return value


class EnforceOverrides(metaclass=EnforceOverridesMeta):
    "Use this as the parent class for your custom classes"
    pass



from overrides import overrides


class Parent:
    def metodi(self, x: int) -> str:
        return f"{x}"


class Child(Parent):
    @overrides
    def metodi(self, x: str) -> int:
        return int(x)



#!/usr/bin/env python

from distutils.core import setup
from setuptools import find_packages  # type: ignore
from os.path import abspath, join, dirname

name = "Mikko Korpela"
# I might be just a little bit too much afraid of those bots..
address = name.lower().replace(" ", ".") + chr(64) + "gmail.com"

desc = "A decorator to automatically detect mismatch when overriding a method."

CURDIR = dirname(abspath(__file__))

with open(join(CURDIR, "README.rst")) as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="overrides",
    version="6.5.0",
    description=desc,
    long_description=LONG_DESCRIPTION,
    author=name,
    author_email=address,
    url="https://github.com/mkorpela/overrides",
    packages=find_packages(),
    package_data={"overrides": ["*.pyi", "py.typed"],},
    include_package_data=True,
    install_requires=['typing;python_version<"3.5"'],
    python_requires=">=3.6",
    license="Apache License, Version 2.0",
    keywords=["override", "inheritence", "OOP"],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



#
#  Copyright 2016 Keunhong Lee
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from types import FunctionType
from typing import Callable, TypeVar, Union

__VERSION__ = "0.1"


_WrappedMethod = TypeVar("_WrappedMethod", bound=Union[FunctionType, Callable])


def final(method: _WrappedMethod) -> _WrappedMethod:
    """Decorator to indicate that the decorated method is finalized and cannot be overridden.
    The decorator code is executed while loading class. Using this method
    should have minimal runtime performance implications.
    Currently, only methods with @overrides are checked.

    How to use:
    from overrides import final

    class SuperClass(object):
        @final
        def method(self):
          return 2

    class SubClass(SuperClass):
        @overrides
        def method(self): #causes an error
            return 1

    :raises AssertionError: if there exists a match in sub classes for the method name
    :return: method
    """
    setattr(method, "__finalized__", True)
    return method
